/*------------------------------------------------------------------------------
 * MDK Middleware - Component ::Network
 * Copyright (c) 2004-2019 Arm Limited (or its affiliates). All rights reserved.
 *------------------------------------------------------------------------------
 * Name:    BSD_Client.c
 * Purpose: BSD socket client example
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "cmsis_os2.h"                  // ::CMSIS:RTOS2
#include "rl_net.h"                     // Keil.MDK-Pro::Network:CORE

#include "BSD_Client.h"
#include "Common_UI.h"

//-------- <<< Use Configuration Wizard in Context Menu >>> -----------------

//   <o>Socket type <0=>Stream <1=>Datagram
//   <i>Stream (TCP) or Datagram (UDP) socket type.
#define SOCK_TYPE       0

//   <o>Remote Port <1-65535>
//   <i>Remote port number.
//   <i>Default: 1001
#define SOCK_PORT       1001

//   <e>Use IPv4
//   <i>Use IPv4 Protocol to connect.
#define IP4_ENABLE      1

//     <s.15>Remote IP Address
//     <i>Remote IPv4 Address in text representation.
//     <i>Default: "192.168.0.100"
#define IP4_ADDR        "192.168.0.100"

//   </e>
//   <e>Use IPv6
//   <i>Use IPv6 Protocol to connect
#define IP6_ENABLE      0

//     <s.40>Remote IP Address
//     <i>Remote IPv6 Address in text representation.
//     <i>Default: "fe80::1c30:6cff:fea2:455e"
#define IP6_ADDR        "fe80::1c30:6cff:fea2:455e"

//   </e>

//------------- <<< end of configuration section >>> -----------------------

#if (!(IP4_ENABLE ^ IP6_ENABLE))
  #error Select only IPv4 or only IPv6!
#endif

#define BLINKLED        0x01
#define SOCKTYPE  ((SOCK_TYPE == 0) ? SOCK_STREAM : SOCK_DGRAM)

/* Main stack size must be multiple of 8 Bytes */
#define APP_MAIN_STK_SZ (1024U)
static uint64_t app_main_stk[APP_MAIN_STK_SZ / 8];
const osThreadAttr_t app_main_attr = {
  .stack_mem  = &app_main_stk[0],
  .stack_size = sizeof(app_main_stk)
};

/* Function declarations */
static void Client (void *arg);
static void AppInitialize (void);
static void print_address (void);

/* Application state */
APP_STATE App;

/* Set application initial states */
static void AppInitialize (void) {

/* Initialize application state structure */
memset (&App, 0x00, sizeof(App));

  UI_Initialize();
#if (IP4_ENABLE)
  UI_SetHello ("  BSD Client IPv4   ");
#endif
#if (IP6_ENABLE)
  UI_SetHello ("  BSD Client IPv6   ");
#endif
}

/* Print local IP addresses */
static void print_address (void) {
#if (IP4_ENABLE)
  uint8_t ip4_addr[NET_ADDR_IP4_LEN];

  if (netIF_GetOption (NET_IF_CLASS_ETH, netIF_OptionIP4_Address,
                                         ip4_addr, sizeof(ip4_addr)) == netOK) {
    /* Print IPv4 network address */
    netIP_ntoa (NET_ADDR_IP4, ip4_addr, App.ip, sizeof(App.ip));

    UI_SetIP (App.ip, NULL);
  }
#else
  uint8_t ip6_addr[NET_ADDR_IP6_LEN];

  if (netIF_GetOption (NET_IF_CLASS_ETH, netIF_OptionIP6_LinkLocalAddress,
                                         ip6_addr, sizeof(ip6_addr)) == netOK) {
    /* Print IPv6 network address (MDK-Pro only) */
    netIP_ntoa(NET_ADDR_IP6, ip6_addr, App.ip, sizeof(App.ip));

    UI_SetIP (NULL, App.ip);
  }
#endif
}

/*----------------------------------------------------------------------------
  Thread 'Client': BSD Client socket process
 *---------------------------------------------------------------------------*/
static __NO_RETURN void Client (void *arg) {
  int32_t  sock, retv;
  uint32_t led_num, led;
  char dbuf[4];

  (void)arg;

  led_num = 1U;

  /* Connect and send LED state */

  while (1) {
    osDelay (2000U);

    print_address();

#if (IP4_ENABLE)
    /* Create IPv4 socket address */
    SOCKADDR_IN  addr;
    addr.sin_family = AF_INET;
    addr.sin_port   = htons(SOCK_PORT);
    if (!inet_pton (AF_INET, IP4_ADDR, &addr.sin_addr)) {
      /* Address configuration error */
      for (;;);
    }

    sock = socket (AF_INET, SOCKTYPE, 0);
    retv = connect (sock, (SOCKADDR *)&addr, sizeof(addr));
#endif

#if (IP6_ENABLE)
    /* Create IPv6 socket address (MDK-Pro only) */
    SOCKADDR_IN6 addr6;
    addr6.sin6_family   = AF_INET6;
    addr6.sin6_port     = htons(SOCK_PORT);
    addr6.sin6_flowinfo = 0;
    if (!inet_pton (AF_INET6, IP6_ADDR, &addr6.sin6_addr)) {
      /* Address configuration error or IPv6 not supported */
      for (;;);
    }

    sock = socket (AF_INET6, SOCKTYPE, 0);
    retv = connect (sock, (SOCKADDR *)&addr6, sizeof(addr6));
#endif

    if (retv < 0) {
      closesocket (sock);
      continue;
    }

    led = 0;
    while (1) {
      // Send data to BSD Server
      dbuf[0] = BLINKLED;
      dbuf[1] = 0x01 << led;
      retv = send (sock, dbuf, 2, 0);
      if (retv < 0) {
        break;
      }
      if (++led >= led_num) {
        led = 0;
      }
      // Blink LED 0
      UI_SetLedOutput (1U);
      osDelay (100U);
      UI_SetLedOutput (0U);
      osDelay (900U);
    }
    closesocket (sock);
  }
}

/*----------------------------------------------------------------------------
  Main Thread 'app_main': Run Network
 *---------------------------------------------------------------------------*/
__NO_RETURN void app_main (void *arg) {
  (void)arg;

  AppInitialize ();
  netInitialize ();
  print_address ();

  // Create networking task
  osThreadNew (Client, NULL, NULL);

  osThreadExit();
}
