/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2019 Arm Limited
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        08. October 2019
 * $Revision:    V1.1
 *
 * Driver:       Driver_USBD2
 * Project:      USB Device Driver for NXP i.MX RT 105x Series
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                  Value
 *   ---------------------                  -----
 *   Connect to hardware via Driver_USBD# = 2
 * --------------------------------------------------------------------------
 * Defines used for driver configuration (at compile time):
 *
 *   USBD_MAX_ENDPOINT_NUM:  defines maximum number of IN/OUT Endpoint pairs 
 *                           that driver will support with Control Endpoint 0
 *                           not included, this value impacts driver memory
 *                           requirements
 *     - default value: 5
 *     - maximum value: 5
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.1
 *    Corrected Endpoint transfer array size
 *  Version 1.0
 *    Initial release
 */

/*! \cond */

#include <stdint.h>
#include <string.h>

#include "Driver_USBD.h"
#include "USB_iMXRT105x.h"

#include "usb.h"
#include "board.h"
#include "usb_phy.h"
#include "usb_device.h"
#include "usb_device_config.h"
#include "usb_device_dci.h"


extern void USB_DeviceEhciIsrFunction(void *deviceHandle);
extern volatile uint8_t USB2_role;
extern volatile uint8_t USB2_state;

static usb_device_handle deviceHandle;

static uint32_t            setup_packet[2];     // Setup packet data
static volatile uint8_t    setup_received;      // Setup packet received

static ARM_USBD_STATE      usbd_state;

#define USBD_IRQn          (USB_OTG2_IRQn)

#define EP_NUM(ep_addr)    (ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_DIR(ep_addr)    ((ep_addr >>  7) & 1U)

// Maximum number of endpoints can be configured in usb_device_config.h
#define USBD_MAX_ENDPOINT_NUM USB_DEVICE_CONFIG_ENDPOINTS

typedef struct {
  uint32_t num;
  uint32_t cnt;
  uint32_t active;
}EP_Transfer;
static EP_Transfer ep_transfer[2][USBD_MAX_ENDPOINT_NUM + 1];

#define CONTROLLER_ID kUSB_ControllerEhci1

// USBD Driver *****************************************************************

#define ARM_USBD_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,1)

// Driver Version
static const ARM_DRIVER_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };
  
// Driver Capabilities
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  1U,   // VBUS Detection
  1U,   // Event VBUS On
  1U    // Event VBUS Off
};

static ARM_USBD_SignalDeviceEvent_t   SignalDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t SignalEndpointEvent;

// Auxiliary functions

static usb_status_t USB_DeviceCallback(usb_device_handle handle, uint32_t event, void *param) {
  usb_status_t error = kStatus_USB_Error;
  uint8_t speed, i, ep_addr;
  
  switch (event) {
    case kUSB_DeviceEventBusReset:
      for (i = 1; i <= USBD_MAX_ENDPOINT_NUM; i++) {
        ep_addr = i;
        USB_DeviceUnstallEndpoint(deviceHandle, ep_addr);
        ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].cnt = 0U;
        ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].active = 0U;
        USB_DeviceCancel(deviceHandle, ep_addr);
        
        ep_addr = ARM_USB_ENDPOINT_DIRECTION_MASK | i;
        USB_DeviceUnstallEndpoint(deviceHandle, ep_addr);
        ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].cnt = 0U;
        ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].active = 0U;
        USB_DeviceCancel(deviceHandle, ep_addr);
      }
      usbd_state.active = 1U;
      usbd_state.vbus   = 1U;
      USB_DeviceGetStatus(deviceHandle, kUSB_DeviceStatusSpeed, &speed);
      if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_RESET); }
      if (speed == USB_SPEED_FULL) {
        usbd_state.speed = ARM_USB_SPEED_FULL;
      } else if (speed == USB_SPEED_HIGH) {
        usbd_state.speed = ARM_USB_SPEED_HIGH;
        if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_HIGH_SPEED); }
      }
      break;
    case kUSB_DeviceEventSuspend:
      usbd_state.active = 0U;
      if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_SUSPEND); }
      break;
    case kUSB_DeviceEventResume:
      usbd_state.active = 1U;
      if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_RESUME); }
      break;
    case kUSB_DeviceEventLPMResume:
      usbd_state.active = 1U;
      if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_RESUME); }
      break;
    case kUSB_DeviceEventDetach:
      usbd_state.active = 0U;
      usbd_state.vbus   = 0U;
      if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_VBUS_OFF); }
      break;
    case kUSB_DeviceEventAttach:
      usbd_state.vbus   = 1U;
      if (SignalDeviceEvent != NULL) { SignalDeviceEvent (ARM_USBD_EVENT_VBUS_ON); }
      break;
  }

  return error;
}

/**
  \fn          usb_status_t USB_DeviceEndpointCb(usb_device_handle handle,
                                                 usb_device_endpoint_callback_message_struct_t *message,
                                                 void *callbackParam)
  \brief       Endpoint callback
  \return      usb_status_t
*/
static usb_status_t USB_DeviceEndpointCb(usb_device_handle handle,
                                             usb_device_endpoint_callback_message_struct_t *message,
                                             void *callbackParam) {
  uint8_t ep_addr = (uint8_t)((uint32_t)callbackParam & 0xFFU);

  if ((ep_addr == 0x00U) && (message->isSetup == 1)) {
    // Setup packet
    memcpy(setup_packet, message->buffer, 8);
    setup_received = 1U;
    SignalEndpointEvent (0U, ARM_USBD_EVENT_SETUP);
  } else if ((ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) != 0) {
    if (ep_transfer[1][EP_NUM(ep_addr)].active != 0U) {
      ep_transfer[1][EP_NUM(ep_addr)].cnt = message->length;
      ep_transfer[1][EP_NUM(ep_addr)].active = 0U;
      SignalEndpointEvent (ep_addr, ARM_USBD_EVENT_IN);
    }
  } else {
    if (ep_transfer[0][EP_NUM(ep_addr)].active != 0U) {
      ep_transfer[0][EP_NUM(ep_addr)].cnt = message->length;
      ep_transfer[0][EP_NUM(ep_addr)].active = 0U;
      SignalEndpointEvent (ep_addr, ARM_USBD_EVENT_OUT);
    }
  }

  return kStatus_USB_Success;
}


// USBD Driver functions

/**
  \fn          ARM_DRIVER_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBD_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_GetCapabilities (void) { return usbd_driver_capabilities; }

/**
  \fn          int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                        ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref execution_status
*/
static int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  usb_phy_config_struct_t phyConfig = {
    BOARD_USB_PHY_D_CAL,
    BOARD_USB_PHY_TXCAL45DP,
    BOARD_USB_PHY_TXCAL45DM 
  };
  
  if ((USB2_state & USBD_DRIVER_INITIALIZED) != 0U) { return ARM_DRIVER_OK; }

  SignalDeviceEvent   = cb_device_event;
  SignalEndpointEvent = cb_endpoint_event;

  USB2_role   =  ARM_USB_ROLE_DEVICE;

  USB2_state  =  USBD_DRIVER_INITIALIZED;

  // Enable clock
  CLOCK_EnableUsbhs1PhyPllClock(kCLOCK_Usbphy480M, 480000000U);
  CLOCK_EnableUsbhs1Clock(kCLOCK_Usb480M, 480000000U);
  USB_EhciPhyInit(CONTROLLER_ID, BOARD_XTAL0_CLK_HZ, &phyConfig);

  // Enable IRQ
  EnableIRQ(USBD_IRQn);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref execution_status
*/
static int32_t USBD_Uninitialize (void) {

  DisableIRQ(USBD_IRQn);
  USB_EhciPhyDeinit(CONTROLLER_ID);

  USB2_role   =  ARM_USB_ROLE_NONE;
  USB2_state &= ~USBD_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBD_PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_DisableIRQ      (USBD_IRQn);                 // Disable interrupt
      NVIC_ClearPendingIRQ (USBD_IRQn);                 // Clear pending interrupt

      USB_DeviceDeinit(deviceHandle);

      usbd_state.active = 0U;

      USB2_state &=  ~USBD_DRIVER_POWERED;
      break;

    case ARM_POWER_FULL:
      if ((USB2_state & USBD_DRIVER_INITIALIZED) == 0U) { return ARM_DRIVER_ERROR; }
      if ((USB2_state & USBD_DRIVER_POWERED)     != 0U) { return ARM_DRIVER_OK; }

      USB_DeviceInit (CONTROLLER_ID, USB_DeviceCallback, &deviceHandle);
      USB2_state |=  USBD_DRIVER_POWERED;               // Set powered flag
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceConnect (void) {

  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB_DeviceRun(deviceHandle);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceDisconnect (void) {

  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB_DeviceStop(deviceHandle);

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBD_STATE USBD_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      Device State \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_DeviceGetState (void) {
  return usbd_state;
}

/**
  \fn          int32_t USBD_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceRemoteWakeup (void) {

  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceSetAddress (uint8_t dev_addr)
  \brief       Set USB Device Address.
  \param[in]   dev_addr  Device Address
  \return      \ref execution_status
*/
static int32_t USBD_DeviceSetAddress (uint8_t dev_addr) {
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB_DeviceSetStatus(deviceHandle, kUSB_DeviceStatusAddress, &dev_addr);
  USB_DeviceSetStatus(deviceHandle, kUSB_DeviceStatusAddress, &dev_addr);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_ReadSetupPacket (uint8_t *setup)
  \brief       Read setup packet received over Control Endpoint.
  \param[out]  setup  Pointer to buffer for setup packet
  \return      \ref execution_status
*/
static int32_t USBD_ReadSetupPacket (uint8_t *setup) {

  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }
  if (setup_received                     == 0U) { return ARM_DRIVER_ERROR; }

  setup_received = 0U;
  memcpy(setup, setup_packet, 8);

  if (setup_received != 0U) {           // If new setup packet was received while this was being read
    return ARM_DRIVER_ERROR;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                               uint8_t  ep_type,
                                               uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type  Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                       uint8_t  ep_type,
                                       uint16_t ep_max_packet_size) {
  uint8_t ep_num;
                                         usb_device_endpoint_init_struct_t     epInit;
  usb_device_endpoint_callback_struct_t epCallback;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)           { return ARM_DRIVER_ERROR; }
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].cnt = 0U;
  ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].active = 0U;

  epInit.maxPacketSize   = ep_max_packet_size;
  epInit.endpointAddress = ep_addr;
  epInit.transferType    = ep_type;
  epInit.zlt             = 0;

  epCallback.callbackFn    = USB_DeviceEndpointCb;
  epCallback.callbackParam = (void *)((uint32_t)ep_addr);

  USB_DeviceInitEndpoint(deviceHandle, &epInit, &epCallback);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointUnconfigure (uint8_t ep_addr) {
  uint8_t  ep_num;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)           { return ARM_DRIVER_ERROR; }
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  USB_DeviceDeinitEndpoint(deviceHandle, ep_addr);

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   stall  Operation
                - \b false Clear
                - \b true Set
  \return      \ref execution_status
*/
static int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall) {
  uint8_t  ep_num;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)           { return ARM_DRIVER_ERROR; }
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  if (stall == true) {
    USB_DeviceStallEndpoint(deviceHandle, ep_addr);
  } else {
    USB_DeviceUnstallEndpoint(deviceHandle, ep_addr);

    ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].cnt = 0U;
    ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].active = 0U;
    USB_DeviceCancel(deviceHandle, ep_addr);
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num)
  \brief       Read data from or Write data to USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[out]  data Pointer to buffer for data to read or with data to write
  \param[in]   num  Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num) {

  if (EP_NUM(ep_addr) > USBD_MAX_ENDPOINT_NUM)  { return ARM_DRIVER_ERROR; }
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep_transfer[1][EP_NUM(ep_addr)].cnt = 0U;
    ep_transfer[1][EP_NUM(ep_addr)].active = 1U;
    USB_DeviceSendRequest(deviceHandle, ep_addr, data, num);
  } else {
    ep_transfer[0][EP_NUM(ep_addr)].cnt = 0U;
    ep_transfer[0][EP_NUM(ep_addr)].active = 1U;
    USB_DeviceRecvRequest(deviceHandle, ep_addr, data, num);
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr)
  \brief       Get result of USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      number of successfully transferred data bytes
*/
static uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr) {

  if (EP_NUM(ep_addr) > USBD_MAX_ENDPOINT_NUM) { return 0U; }

  return (ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].cnt);
}

/**
  \fn          int32_t USBD_EndpointTransferAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransferAbort (uint8_t ep_addr) {
  uint8_t  ep_num;

  ep_num = EP_NUM(ep_addr);
  if (ep_num > USBD_MAX_ENDPOINT_NUM)           { return ARM_DRIVER_ERROR; }
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return ARM_DRIVER_ERROR; }

  ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].cnt = 0U;
  ep_transfer[(ep_addr >> 7) & 1U][EP_NUM(ep_addr)].active = 0U;
  USB_DeviceCancel(deviceHandle, ep_addr);

  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t USBD_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_GetFrameNumber (void) {
  uint16_t frame;
  if ((USB2_state & USBD_DRIVER_POWERED) == 0U) { return 0U; }

  USB_DeviceGetStatus (deviceHandle, kUSB_DeviceStatusSynchFrame, (void *)(&frame));
  return (frame);
}

void USBD2_IRQ (void)
{
    USB_DeviceEhciIsrFunction(deviceHandle);
}

ARM_DRIVER_USBD Driver_USBD2 = {
  USBD_GetVersion,
  USBD_GetCapabilities,
  USBD_Initialize,
  USBD_Uninitialize,
  USBD_PowerControl,
  USBD_DeviceConnect,
  USBD_DeviceDisconnect,
  USBD_DeviceGetState,
  USBD_DeviceRemoteWakeup,
  USBD_DeviceSetAddress,
  USBD_ReadSetupPacket,
  USBD_EndpointConfigure,
  USBD_EndpointUnconfigure,
  USBD_EndpointStall,
  USBD_EndpointTransfer,
  USBD_EndpointTransferGetResult,
  USBD_EndpointTransferAbort,
  USBD_GetFrameNumber
};

/*! \endcond */
