/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2018 Arm Limited
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        24. April 2018
 * $Revision:    V1.0
 *
 * Driver:       Driver_USBH1_HCI
 * Project:      USB Host 1 HCI Controller (EHCI) Driver for 
 *               NXP i.MX RT 105x Series
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                  Value
 *   ---------------------                  -----
 *   Connect to hardware via Driver_USBH# = 1
 *   USB Host controller interface        = EHCI
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.0
 *    Initial release
 */

/*! \cond */

#include "Driver_USBH.h"
#include "USB_iMXRT105x.h"

#include "fsl_device_registers.h"
#include "fsl_clock.h"

extern volatile uint8_t USB1_role;
extern volatile uint8_t USB1_state;


// USBH EHCI Driver ************************************************************

#define ARM_USBH_EHCI_DRIVER_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,0)

// Driver Version
static const ARM_DRIVER_VERSION usbh_ehci_driver_version = { ARM_USBH_API_VERSION, ARM_USBH_EHCI_DRIVER_VERSION };

// Driver Capabilities
static const ARM_USBH_HCI_CAPABILITIES usbh_ehci_driver_capabilities = {
  0x0001U       // Root HUB available Ports Mask
};

static ARM_USBH_HCI_Interrupt_t EHCI_IRQ;

// USBH EHCI Driver functions

/**
  \fn          ARM_DRIVER_VERSION USBH_HCI_GetVersion (void)
  \brief       Get USB Host HCI (OHCI/EHCI) driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBH_HCI_GetVersion (void) { return usbh_ehci_driver_version; }

/**
  \fn          ARM_USBH_HCI_CAPABILITIES USBH_HCI_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBH_HCI_CAPABILITIES
*/
static ARM_USBH_HCI_CAPABILITIES USBH_HCI_GetCapabilities (void) { return usbh_ehci_driver_capabilities; }

/**
  \fn          int32_t USBH_HCI_Initialize (ARM_USBH_HCI_Interrupt_t *cb_interrupt)
  \brief       Initialize USB Host HCI (OHCI/EHCI) Interface.
  \param[in]   cb_interrupt Pointer to Interrupt Handler Routine
  \return      \ref execution_status
*/
static int32_t USBH_HCI_Initialize (ARM_USBH_HCI_Interrupt_t cb_interrupt) {

  if ((USB1_state & USBH_DRIVER_INITIALIZED) != 0U) { return ARM_DRIVER_OK; }

  EHCI_IRQ    =  cb_interrupt;

  USB1_role   =  ARM_USB_ROLE_HOST;

  USB1_state  =  USBH_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_HCI_Uninitialize (void)
  \brief       De-initialize USB Host HCI (OHCI/EHCI) Interface.
  \return      \ref execution_status
*/
static int32_t USBH_HCI_Uninitialize (void) {

  USB1_role   =  ARM_USB_ROLE_NONE;
  USB1_state &= ~USBH_DRIVER_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_HCI_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Host HCI (OHCI/EHCI) Interface Power.
  \param[in]   state Power state
  \return      \ref execution_status
*/
static int32_t USBH_HCI_PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_DisableIRQ      (USB_OTG1_IRQn);             // Disable interrupt
      NVIC_ClearPendingIRQ (USB_OTG1_IRQn);             // Clear pending interrupt
      USB1_state &= ~USBH_DRIVER_POWERED;               // Clear powered flag
      CLOCK_DisableUsbhs0PhyPllClock();
      break;

    case ARM_POWER_FULL:
      if ((USB1_state & USBH_DRIVER_INITIALIZED) == 0U) { return ARM_DRIVER_ERROR; }
      if ((USB1_state & USBH_DRIVER_POWERED)     != 0U) { return ARM_DRIVER_OK; }

      if (CLOCK_EnableUsbhs0PhyPllClock(kCLOCK_Usbphy480M, 480000000U) == false) {
        return ARM_DRIVER_ERROR;
      }

      USB1_state |=  USBH_DRIVER_POWERED;               // Set powered flag
      NVIC_EnableIRQ   (USB_OTG1_IRQn);                 // Enable interrupt
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBH_HCI_PortVbusOnOff (uint8_t port, bool vbus)
  \brief       USB Host HCI (OHCI/EHCI) Root HUB Port VBUS on/off.
  \param[in]   port  Root HUB Port Number
  \param[in]   vbus
                - \b false VBUS off
                - \b true  VBUS on
  \return      \ref execution_status
*/
static int32_t USBH_HCI_PortVbusOnOff (uint8_t port, bool power) {
  // VBUS is always provided on board, it cannot be turned off

  if (((1U << port) & usbh_ehci_driver_capabilities.port_mask) == 0U) { return ARM_DRIVER_ERROR; }
  return ARM_DRIVER_OK;
}

/**
  \fn          void USBH1_IRQ (void)
  \brief       USB1 Host Interrupt Routine (IRQ).
*/
void USBH1_IRQ (void) {
  uint32_t usbsts, portsc1;

  // Special handling
  usbsts = USB1->USBSTS;
  if ((usbsts & USB_USBSTS_PCI_MASK) != 0U) {           // If Port Change Detect is active
    portsc1 = USB1->PORTSC1;
    if ((portsc1 & USB_PORTSC1_SUSP_MASK)!= 0U) {       // If Suspend is active disable HS disconnect detector
      USBPHY1->CTRL &= ~USBPHY_CTRL_ENHOSTDISCONDETECT_MASK;
    } else if ( ((portsc1 & USB_PORTSC1_CCS_MASK) != 0U)                             &&
               (((portsc1 & USB_PORTSC1_PSPD_MASK) >> USB_PORTSC1_PSPD_SHIFT) == 2U) &&
               ( (portsc1 & USB_PORTSC1_FPR_MASK)                             == 0U)) {
      // If Current Connect Status is active and HS Device Connected and no forced suspend, 
      // enable HS disconnect detector
      USBPHY1->CTRL |=  USBPHY_CTRL_ENHOSTDISCONDETECT_MASK;
    } else {                    // else disable HS disconnect detector
      USBPHY1->CTRL &= ~USBPHY_CTRL_ENHOSTDISCONDETECT_MASK;
    }
  }

  EHCI_IRQ();
}

ARM_DRIVER_USBH_HCI Driver_USBH1_HCI = {
  USBH_HCI_GetVersion,
  USBH_HCI_GetCapabilities,
  USBH_HCI_Initialize,
  USBH_HCI_Uninitialize,
  USBH_HCI_PowerControl,
  USBH_HCI_PortVbusOnOff
};

/*! \endcond */
